///////////////////////////////////////////////////////////////////////////////
// ControllerMain.cpp
// ==================
// Derived Controller class for main window
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2023-04-12
// UPDATED: 2023-04-20
///////////////////////////////////////////////////////////////////////////////

#include "ControllerMain.h"
#include "resource.h"
#include "Log.h"

using namespace Win;

// low-level keyboard/mouse event hooks
HHOOK hookKeyboard = 0;
HHOOK hookMouse = 0;
HWND handleWindow = 0;      // handle for this window
Win::ViewForm* viewPtr = 0; // ptr to view of form dialog
LRESULT CALLBACK hookKeyboardProcedure(int code, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK hookMouseProcedure(int code, WPARAM wParam, LPARAM lParam);



// handle events(messages) on all child windows that belong to the parent window.
// For example, close all child windows when the parent got WM_CLOSE message.
// lParam can be used to specify a event or message.
bool CALLBACK enumerateChildren(HWND childHandle, LPARAM lParam);

///////////////////////////////////////////////////////////////////////////////
// enumerate all child windows
///////////////////////////////////////////////////////////////////////////////
bool CALLBACK enumerateChildren(HWND handle, LPARAM lParam)
{
    if(lParam == WM_CLOSE)
    {
        ::SendMessage(handle, WM_CLOSE, 0, 0);  // close child windows
    }

    return true;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_CLOSE message
///////////////////////////////////////////////////////////////////////////////
LRESULT ControllerMain::close()
{
    // close all child windows first
    ::EnumChildWindows(handle, (WNDENUMPROC)enumerateChildren, (LPARAM)WM_CLOSE);

    log(L"Main window is closed");
    ::DestroyWindow(handle);                    // destroy itself
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_DESTROY message
///////////////////////////////////////////////////////////////////////////////
LRESULT ControllerMain::destroy()
{
    // uninstall event hooks
    if(::UnhookWindowsHookEx(hookKeyboard))
        Win::log("Uninstalled low-level keyboard event hook successfully.");
    else
        Win::log("[ERROR] Failed to uninstall keyboard event hook.");

    if(::UnhookWindowsHookEx(hookMouse))
        Win::log("Uninstalled low-level mouse event hook successfully.");
    else
        Win::log("[ERROR] Failed to uninstall keyboard event hook.");

    ::PostQuitMessage(0);       // exit the message loop
    log(L"Main window is destroyed");
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_CREATE message
///////////////////////////////////////////////////////////////////////////////
LRESULT ControllerMain::create()
{
    return 0;
}




///////////////////////////////////////////////////////////////////////////////
// handle WM_SIZE notification
// Note that the input param, width and height is for client area only.
// It excludes non-client area.
///////////////////////////////////////////////////////////////////////////////
LRESULT ControllerMain::size(int width, int height, WPARAM type)
{
    log(L"Main window is resized: %dx%d", width, height);
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_COMMAND notification
///////////////////////////////////////////////////////////////////////////////
LRESULT ControllerMain::command(int id, int command, LPARAM message)
{
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// install low-level keyboard/mouse event hooks
// remember the view component of form dialog
///////////////////////////////////////////////////////////////////////////////
void ControllerMain::initEventHooks(ViewForm* view)
{
    this->viewForm = view;

    // install low-level keyboard/mouse hooks to detect events
    handleWindow = handle;  // this window handle
    viewPtr = viewForm;
    HINSTANCE instance = ::GetModuleHandle(0);  // get app instance ID
    hookKeyboard = ::SetWindowsHookEx(WH_KEYBOARD_LL, hookKeyboardProcedure, instance, 0);
    hookMouse = ::SetWindowsHookEx(WH_MOUSE_LL, hookMouseProcedure, instance, 0);
    if(hookKeyboard)
        Win::log("Installed low-level keyboard event hook from ControllerMain.");
    else
        Win::log("[ERROR] Failed to install keyboard event hook from ControllerMain.");
    if(hookMouse)
        Win::log("Installed low-level mouse event hook from ControllerMain.");
    else
        Win::log("[ERROR] Failed to install mouse event hook from ControllerMain.");
}



//=============================================================================
///////////////////////////////////////////////////////////////////////////////
// hook for low-level keyboard events
// code < 0: must return it without processing it
// code = 0: (HC_ACTION),contains keystroke message in wParam & lParam
// code = 3: (HC_NOREMOVE), contains keystroke message but not removed
///////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK hookKeyboardProcedure(int code, WPARAM wParam, LPARAM lParam)
{
    // do not process if code < 0
    if(code >= 0)
    {
        // wParam contains keyboard event message name
        // lParam contains ptr to KBDLLHOOKSTRUCT
        KBDLLHOOKSTRUCT* khs = (KBDLLHOOKSTRUCT*)lParam;
        //DWORD keyCode = khs->vkCode;    // virtual key code, 1 ~ 254
        //DWORD scanCode = khs->scanCode; // hardware scan code for key
        //DWORD flags = khs->flags;       // extended flags
        //DWORD time = khs->time;         // timestamp
        //ULONG_PTR extraInfo = khs->dwExtraInfo;// ptr to additional info

        switch(wParam)
        {
        case WM_KEYDOWN:
            viewPtr->updateKeyDown((int)khs->vkCode);   // pass key code
            break;

        case WM_KEYUP:
            viewPtr->updateKeyUp((int)khs->vkCode);     // pass key code
            break;
        }
    }

    // chain to the next hook procedures
    return ::CallNextHookEx(hookKeyboard, code, wParam, lParam);
}



///////////////////////////////////////////////////////////////////////////////
// hook procedure for low-level mouse events
// code < 0: must return it without processing it
// code = 0: (HC_ACTION),contains mouse message in wParam & lParam
///////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK hookMouseProcedure(int code, WPARAM wParam, LPARAM lParam)
{
    if(code >= 0)
    {
        // wParam contains mouse event message name
        // lParam contains ptr to MSLLHOOKSTRUCT
        MSLLHOOKSTRUCT* mhs = (MSLLHOOKSTRUCT*)lParam;
        //POINT point = mhs->pt;          // mouse position in screen space
        //DWORD data = mhs->mouseData;    // wheel delta for WM_MOUSEWHEEL
        //DWORD flags = mhs->flags;       // extended flags
        //DWORD time = mhs->time;         // timestamp
        //ULONG_PTR extraInfo = mhs->dwExtraInfo;// ptr to additional info

        POINT point;    // to store mouse position in window space

        switch(wParam)
        {
        case WM_LBUTTONDOWN:
            viewPtr->updateMouseDown(1);
            break;
        case WM_MBUTTONDOWN:
            viewPtr->updateMouseDown(2);
            break;
        case WM_RBUTTONDOWN:
            viewPtr->updateMouseDown(3);
            break;

        case WM_LBUTTONUP:
            viewPtr->updateMouseUp(1);
            break;
        case WM_MBUTTONUP:
            viewPtr->updateMouseUp(2);
            break;
        case WM_RBUTTONUP:
            viewPtr->updateMouseUp(3);
            break;

        case WM_MOUSEWHEEL:
            //Win::log(L"wheel: %d", GET_WHEEL_DELTA_WPARAM(mhs->mouseData));
            viewPtr->updateMouseWheel(GET_WHEEL_DELTA_WPARAM(mhs->mouseData));
            break;

        case WM_MOUSEMOVE:
            // convert screen coord to window coord
            point.x = mhs->pt.x;
            point.y = mhs->pt.y;
            ::ScreenToClient(handleWindow, &point);
            viewPtr->updateMousePosition((int)point.x, (int)point.y);
            //Win::log("POS: (%d,%d), (%d,%d)", mhs->pt.x, mhs->pt.y, point.x, point.y);
            break;
        }
    }

    // chain to the next hook procedures
    return ::CallNextHookEx(hookMouse, code, wParam, lParam);
}
